/**
 * Copyright (c) 2014 - 2020, Nordic Semiconductor ASA
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form, except as embedded into a Nordic
 *    Semiconductor ASA integrated circuit in a product or a software update for
 *    such product, must reproduce the above copyright notice, this list of
 *    conditions and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. Neither the name of Nordic Semiconductor ASA nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * 4. This software, with or without modification, must only be used with a
 *    Nordic Semiconductor ASA integrated circuit.
 *
 * 5. Any software provided in binary form under this license must not be reverse
 *    engineered, decompiled, modified and/or disassembled.
 *
 * THIS SOFTWARE IS PROVIDED BY NORDIC SEMICONDUCTOR ASA "AS IS" AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */
/** @file
 *
 * @defgroup ble_sdk_uart_over_ble_main main.c
 * @{
 * @ingroup  ble_sdk_app_nus_eval
 * @brief    UART over BLE application main file.
 *
 * This file contains the source code for a sample application that uses the Nordic UART service.
 * This application uses the @ref srvlib_conn_params module.
 */

#include "app_timer.h"
#include "app_util_platform.h"
#include "ble_advdata.h"
#include "ble_advertising.h"
#include "ble_conn_params.h"
#include "ble_hci.h"
#include "ble_nus.h"
#include "bsp_btn_ble.h"
#include "nordic_common.h"
#include "nrf.h"
#include "nrf_ble_gatt.h"
#include "nrf_ble_qwr.h"
#include "nrf_delay.h"
#include "nrf_libuarte_async.h"
#include "nrf_pwr_mgmt.h"
#include "nrf_sdh.h"
#include "nrf_sdh_ble.h"
#include "nrf_sdh_soc.h"
#include <stdint.h>
#include <string.h>

#include "nrf_drv_clock.h"

#include "ant_bpwr.h"
#include "ant_interface.h"
#include "ant_key_manager.h"
#include "ant_state_indicator.h"

#include "nrf_log.h"
#include "nrf_log_ctrl.h"
#include "nrf_log_default_backends.h"

#define APP_BLE_CONN_CFG_TAG 1 /**< A tag identifying the SoftDevice BLE configuration. */

#define DEVICE_NAME "nRF_libUARTE"                       /**< Name of device. Will be included in the advertising data. */
#define NUS_SERVICE_UUID_TYPE BLE_UUID_TYPE_VENDOR_BEGIN /**< UUID type for the Nordic UART Service (vendor specific). */

#define APP_BLE_OBSERVER_PRIO 3 /**< Application's BLE observer priority. You shouldn't need to modify this value. */

#define APP_ADV_INTERVAL 64 /**< The advertising interval (in units of 0.625 ms. This value corresponds to 40 ms). */

#define APP_ADV_DURATION 18000 /**< The advertising duration (180 seconds) in units of 10 milliseconds. */

#define MIN_CONN_INTERVAL MSEC_TO_UNITS(20, UNIT_1_25_MS)    /**< Minimum acceptable connection interval (20 ms), Connection interval uses 1.25 ms units. */
#define MAX_CONN_INTERVAL MSEC_TO_UNITS(75, UNIT_1_25_MS)    /**< Maximum acceptable connection interval (75 ms), Connection interval uses 1.25 ms units. */
#define SLAVE_LATENCY 0                                      /**< Slave latency. */
#define CONN_SUP_TIMEOUT MSEC_TO_UNITS(4000, UNIT_10_MS)     /**< Connection supervisory timeout (4 seconds), Supervision Timeout uses 10 ms units. */
#define FIRST_CONN_PARAMS_UPDATE_DELAY APP_TIMER_TICKS(5000) /**< Time from initiating event (connect or start of notification) to first time sd_ble_gap_conn_param_update is called (5 seconds). */
#define NEXT_CONN_PARAMS_UPDATE_DELAY APP_TIMER_TICKS(30000) /**< Time between each call to sd_ble_gap_conn_param_update after the first call (30 seconds). */
#define MAX_CONN_PARAMS_UPDATE_COUNT 3                       /**< Number of attempts before giving up the connection parameter negotiation. */

#define DEAD_BEEF 0xDEADBEEF

#define NRF_IF0_2_CTRL 13
#define NRF_IF3_5_CTRL 24

NRF_LIBUARTE_ASYNC_DEFINE(libuarte, 0, 1, 2, NRF_LIBUARTE_PERIPHERAL_NOT_USED, 255, 4);
//NRF_LIBUARTE_ASYNC_DEFINE(libuarte, 0, 0, 0, NRF_LIBUARTE_PERIPHERAL_NOT_USED, 255, 3);


BLE_NUS_DEF(m_nus, NRF_SDH_BLE_TOTAL_LINK_COUNT); /**< BLE NUS service instance. */
NRF_BLE_GATT_DEF(m_gatt);                         /**< GATT module instance. */
NRF_BLE_QWR_DEF(m_qwr);                           /**< Context for the Queued Write module.*/
BLE_ADVERTISING_DEF(m_advertising);               /**< Advertising module instance. */

static uint16_t m_conn_handle = BLE_CONN_HANDLE_INVALID;               /**< Handle of the current connection. */
static uint16_t m_ble_nus_max_data_len = BLE_GATT_ATT_MTU_DEFAULT - 3; /**< Maximum length of data (in bytes) that can be transmitted to the peer by the Nordic UART service module. */
static ble_uuid_t m_adv_uuids[] =                                      /**< Universally unique service identifier. */
    {
        {BLE_UUID_NUS_SERVICE, NUS_SERVICE_UUID_TYPE}};

/* Random Counters for debug */
uint8_t sc = 0;
/** @snippet [ANT BPWR RX Instance] */

void ant_evt_handler(ant_evt_t *p_ant_evt, void *p_context);

NRF_SDH_ANT_OBSERVER(m_ant_observer, APP_ANT_OBSERVER_PRIO,
    ant_evt_handler, NULL);

typedef struct
{
  uint8_t update_event_count;
  uint8_t pedal_power;
  uint8_t reserved;
  uint8_t accumulated_power[2];
  uint8_t instantaneous_power[2];
} ant_bpwr_page16_data_layout_t;

uint8_t uart_sent[64];

void uart_tx() {
  ret_code_t err_code;

  //static uint8_t text[] = "ble_app_libUARTE example started.\n";
  //static uint8_t text[] = { 0x78, 0xFF, 0x00, 0x19, 0x00, 0x00, 0x62, 0x6C, 0x65, 0x5F, 0x61, 0x70, 0x70, 0x0A  };
  //static uint8_t text[] = { 0x9c, 0xb9, 0x73, 0x1b, 0x10, 0x47, 0x00 };
  static uint8_t text[] = {0xac, 0xc8, 0x28, 0x57, 0x12, 0x4a, 0x00};
  static uint8_t text_size = sizeof(text);

  for (int i = 0; i < text_size; i++) {
    printf("%02X ", text[i]);
  }
  printf(" (%d)\n", sc);
  printf("Send: %d\n", text_size);

  ant_bpwr_page16_data_layout_t const *p_incoming_data =
      (ant_bpwr_page16_data_layout_t *)text;

  printf("update_event_count:  %u\n", p_incoming_data->update_event_count);
  printf("pedal_power:         %u\n", p_incoming_data->pedal_power);
  printf("reserved:            %u\n", p_incoming_data->reserved);
  printf("accumulated_power:   %u\n", uint16_decode(p_incoming_data->accumulated_power));
  printf("instantaneous_power: %u\n", uint16_decode(p_incoming_data->instantaneous_power));

  //err_code = nrf_libuarte_async_tx(&libuarte, text, text_size);

  //APP_ERROR_CHECK(err_code);
}

/**@brief Function for handling a ANT stack event.
 *
 * @param[in] p_ant_evt  ANT stack event.
 * @param[in] p_context  Context.
 */
void ant_evt_handler(ant_evt_t *p_ant_evt, void *p_context) {

  ret_code_t err_code;

  if (p_ant_evt->channel == ANT_CHANNEL_NUM) {
    switch (p_ant_evt->event) {
    case EVENT_RX:
        // Really strange, need this (;) or error: see:
        // https://stackoverflow.com/questions/18496282/why-do-i-get-a-label-can-only-be-part-of-a-statement-and-a-declaration-is-not-a
        ;
      //unsigned char *charPtr = (unsigned char *)p_ant_evt;
      ANT_MESSAGE *p_ant_message = (ANT_MESSAGE *)&p_ant_evt->message;

      // Could use ANT_STANDARD_DATA_PAYLOAD_SIZE for the size ?

      unsigned char *charPtr = (unsigned char *)p_ant_message->ANT_MESSAGE_aucPayload;
      if (p_ant_message->ANT_MESSAGE_aucPayload[0] != 0x10) {
        break;
      }

      if (uart_sent[(uint8_t)p_ant_message->ANT_MESSAGE_aucPayload[0]] == p_ant_message->ANT_MESSAGE_aucPayload[1]) {
        //NRF_LOG_ERROR("Already Sent: %u", p_ant_message->ANT_MESSAGE_aucPayload[1])
        break;
      }

      uart_sent[(uint8_t)p_ant_message->ANT_MESSAGE_aucPayload[0]] = p_ant_message->ANT_MESSAGE_aucPayload[1];
      //NRF_LOG_ERROR("Size of payload: %u %u", sizeof(p_ant_message->ANT_MESSAGE_aucPayload), p_ant_message->ANT_MESSAGE_ucSize);

      //static uint8_t text[] = { 0x10, 0xac, 0xc8, 0x28, 0x57, 0x12, 0x4a, 0x00 };
      static uint8_t text[ANT_STANDARD_DATA_PAYLOAD_SIZE];
      static uint8_t text_size = sizeof(text);

      for (int i = 0; i < ANT_STANDARD_DATA_PAYLOAD_SIZE; i++) {
        printf("%02X ", p_ant_message->ANT_MESSAGE_aucPayload[i]);
        text[i] = p_ant_message->ANT_MESSAGE_aucPayload[i];
      }
      printf(" (%d)\n", sc);
      sc++;

      //err_code = nrf_libuarte_async_tx(&libuarte, (uint8_t *)&p_ant_message->ANT_MESSAGE_aucPayload, ANT_STANDARD_DATA_PAYLOAD_SIZE);
      err_code = nrf_libuarte_async_tx(&libuarte, text, text_size);
      APP_ERROR_CHECK(err_code);

      if (p_ant_evt->message.ANT_MESSAGE_ucMesgID == MESG_BROADCAST_DATA_ID || p_ant_evt->message.ANT_MESSAGE_ucMesgID == MESG_ACKNOWLEDGED_DATA_ID || p_ant_evt->message.ANT_MESSAGE_ucMesgID == MESG_BURST_DATA_ID) {
        // LED 2 is inverted for 100 msec
        err_code = bsp_indication_set(BSP_INDICATE_RCV_OK);
        APP_ERROR_CHECK(err_code);
      }
      break;

    default:
      break;
    }
  }
}

/**@brief Function for assert macro callback.
 *
 * @details This function will be called in case of an assert in the SoftDevice.
 *
 * @warning This handler is an example only and does not fit a final product. You need to analyse
 *          how your product is supposed to react in case of Assert.
 * @warning On assert from the SoftDevice, the system can only recover on reset.
 *
 * @param[in] line_num    Line number of the failing ASSERT call.
 * @param[in] p_file_name File name of the failing ASSERT call.
 */
void assert_nrf_callback(uint16_t line_num, const uint8_t *p_file_name) {
  app_error_handler(DEAD_BEEF, line_num, p_file_name);
}

/**@brief Function for initializing the timer module.
 */
static void timers_init(void) {
  ret_code_t err_code = app_timer_init();
  APP_ERROR_CHECK(err_code);
}

/**@brief Function for handling Queued Write Module errors.
 *
 * @details A pointer to this function will be passed to each service which may need to inform the
 *          application about an error.
 *
 * @param[in]   nrf_error   Error code containing information about what went wrong.
 */
static void nrf_qwr_error_handler(uint32_t nrf_error) {
  APP_ERROR_HANDLER(nrf_error);
}

/**@brief Function for handling the data from the Nordic UART Service.
 *
 * @details This function will process the data received from the Nordic UART BLE Service and send
 *          it to the UART module.
 *
 * @param[in] p_evt       Nordic UART Service event.
 */
/**@snippet [Handling the data received over BLE] */
static void nus_data_handler(ble_nus_evt_t *p_evt) {

  if (p_evt->type == BLE_NUS_EVT_RX_DATA) {
    uint32_t err_code;

    NRF_LOG_DEBUG("Received data from BLE NUS. Writing data on UART.");
    NRF_LOG_HEXDUMP_DEBUG(p_evt->params.rx_data.p_data, p_evt->params.rx_data.length);

    uint8_t uart_string[BLE_NUS_MAX_DATA_LEN + 2] = "";
    memcpy(uart_string, p_evt->params.rx_data.p_data, p_evt->params.rx_data.length);
    uart_string[p_evt->params.rx_data.length] = '\r';
    uart_string[p_evt->params.rx_data.length + 1] = '\n';

    err_code = nrf_libuarte_async_tx(&libuarte, uart_string, p_evt->params.rx_data.length + 2);
    APP_ERROR_CHECK(err_code);
  }
}
/**@snippet [Handling the data received over BLE] */


/**@brief Function for handling an event from the Connection Parameters Module.
 *
 * @details This function will be called for all events in the Connection Parameters Module
 *          which are passed to the application.
 *
 * @note All this function does is to disconnect. This could have been done by simply setting
 *       the disconnect_on_fail config parameter, but instead we use the event handler
 *       mechanism to demonstrate its use.
 *
 * @param[in] p_evt  Event received from the Connection Parameters Module.
 */
static void on_conn_params_evt(ble_conn_params_evt_t *p_evt) {
  uint32_t err_code;

  if (p_evt->evt_type == BLE_CONN_PARAMS_EVT_FAILED) {
    err_code = sd_ble_gap_disconnect(m_conn_handle, BLE_HCI_CONN_INTERVAL_UNACCEPTABLE);
    APP_ERROR_CHECK(err_code);
  }
}

/**@brief Function for handling errors from the Connection Parameters module.
 *
 * @param[in] nrf_error  Error code containing information about what went wrong.
 */
static void conn_params_error_handler(uint32_t nrf_error) {
  APP_ERROR_HANDLER(nrf_error);
}

/**@brief Function for putting the chip into sleep mode.
 *
 * @note This function will not return.
 */
static void sleep_mode_enter(void) {
  uint32_t err_code = bsp_indication_set(BSP_INDICATE_IDLE);
  APP_ERROR_CHECK(err_code);

  // Prepare wakeup buttons.
  err_code = bsp_btn_ble_sleep_mode_prepare();
  APP_ERROR_CHECK(err_code);

  // Go to system-off mode (this function will not return; wakeup will cause a reset).
  err_code = sd_power_system_off();
  APP_ERROR_CHECK(err_code);
}

/**@brief Function for handling advertising events.
 *
 * @details This function will be called for advertising events which are passed to the application.
 *
 * @param[in] ble_adv_evt  Advertising event.
 */
static void on_adv_evt(ble_adv_evt_t ble_adv_evt) {
  uint32_t err_code;

  switch (ble_adv_evt) {
  case BLE_ADV_EVT_FAST:
    err_code = bsp_indication_set(BSP_INDICATE_ADVERTISING);
    APP_ERROR_CHECK(err_code);
    break;
  case BLE_ADV_EVT_IDLE:
    sleep_mode_enter();
    break;
  default:
    break;
  }
}

/**@brief Function for handling BLE events.
 *
 * @param[in]   p_ble_evt   Bluetooth stack event.
 * @param[in]   p_context   Unused.

static void ble_evt_handler(ble_evt_t const *p_ble_evt, void *p_context) {
  uint32_t err_code;

  switch (p_ble_evt->header.evt_id) {
  case BLE_GAP_EVT_CONNECTED:
    NRF_LOG_INFO("Connected");
    err_code = bsp_indication_set(BSP_INDICATE_CONNECTED);
    APP_ERROR_CHECK(err_code);
    m_conn_handle = p_ble_evt->evt.gap_evt.conn_handle;
    err_code = nrf_ble_qwr_conn_handle_assign(&m_qwr, m_conn_handle);
    APP_ERROR_CHECK(err_code);
    break;

  case BLE_GAP_EVT_DISCONNECTED:
    NRF_LOG_INFO("Disconnected");
    // LED indication will be changed when advertising starts.
    m_conn_handle = BLE_CONN_HANDLE_INVALID;
    break;

  case BLE_GAP_EVT_PHY_UPDATE_REQUEST: {
    NRF_LOG_DEBUG("PHY update request.");
    ble_gap_phys_t const phys =
        {
            .rx_phys = BLE_GAP_PHY_AUTO,
            .tx_phys = BLE_GAP_PHY_AUTO,
        };
    err_code = sd_ble_gap_phy_update(p_ble_evt->evt.gap_evt.conn_handle, &phys);
    APP_ERROR_CHECK(err_code);
  } break;

  case BLE_GAP_EVT_SEC_PARAMS_REQUEST:
    // Pairing not supported
    err_code = sd_ble_gap_sec_params_reply(m_conn_handle, BLE_GAP_SEC_STATUS_PAIRING_NOT_SUPP, NULL, NULL);
    APP_ERROR_CHECK(err_code);
    break;

  case BLE_GATTS_EVT_SYS_ATTR_MISSING:
    // No system attributes have been stored.
    err_code = sd_ble_gatts_sys_attr_set(m_conn_handle, NULL, 0, 0);
    APP_ERROR_CHECK(err_code);
    break;

  case BLE_GATTC_EVT_TIMEOUT:
    // Disconnect on GATT Client timeout event.
    err_code = sd_ble_gap_disconnect(p_ble_evt->evt.gattc_evt.conn_handle,
        BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
    APP_ERROR_CHECK(err_code);
    break;

  case BLE_GATTS_EVT_TIMEOUT:
    // Disconnect on GATT Server timeout event.
    err_code = sd_ble_gap_disconnect(p_ble_evt->evt.gatts_evt.conn_handle,
        BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
    APP_ERROR_CHECK(err_code);
    break;

  default:
    // No implementation needed.
    break;
  }
}
*/

/**
 * @brief Function for ANT stack initialization.
 */
void ant_stack_init(void) {

  ret_code_t err_code = nrf_sdh_enable_request();
  APP_ERROR_CHECK(err_code);

  ASSERT(nrf_sdh_is_enabled());

  err_code = nrf_sdh_ant_enable();
  APP_ERROR_CHECK(err_code);

  err_code = ant_plus_key_set(ANTPLUS_NETWORK_NUM);
  APP_ERROR_CHECK(err_code);
}

/**@brief Function for setting up the ANT module to be ready for RX broadcast.
 */
static void ant_channel_rx_broadcast_setup(void) {
  NRF_LOG_INFO("ant_channel_rx_broadcast_setup.");
  ant_channel_config_t broadcast_channel_config =
      {
          //.channel_number    = BROADCAST_CHANNEL_NUMBER,
          .channel_number = ANT_CHANNEL_NUM,
          .channel_type = CHANNEL_TYPE_SLAVE,
          .ext_assign = 0x00,
          .rf_freq = RF_FREQ,
          //.rf_freq           = 0x39,
          .transmission_type = CHAN_ID_TRANS_TYPE,
          .device_type = CHAN_ID_DEV_TYPE,
          .device_number = CHAN_ID_DEV_NUM,
          .channel_period = CHAN_PERIOD,
          .network_number = ANTPLUS_NETWORK_NUM};

  ret_code_t err_code = ant_channel_init(&broadcast_channel_config);
  APP_ERROR_CHECK(err_code);

  // Open channel.
  err_code = sd_ant_channel_open(ANT_CHANNEL_NUM);
  APP_ERROR_CHECK(err_code);
}

/**@brief Function for handling events from the GATT library. 
void gatt_evt_handler(nrf_ble_gatt_t *p_gatt, nrf_ble_gatt_evt_t const *p_evt) {
  if ((m_conn_handle == p_evt->conn_handle) && (p_evt->evt_id == NRF_BLE_GATT_EVT_ATT_MTU_UPDATED)) {
    m_ble_nus_max_data_len = p_evt->params.att_mtu_effective - OPCODE_LENGTH - HANDLE_LENGTH;
    NRF_LOG_INFO("Data len is set to 0x%X(%d)", m_ble_nus_max_data_len, m_ble_nus_max_data_len);
  }
  NRF_LOG_DEBUG("ATT MTU exchange completed. central 0x%x peripheral 0x%x",
      p_gatt->att_mtu_desired_central,
      p_gatt->att_mtu_desired_periph);
}
*/


/**@brief Function for handling events from the BSP module.
 *
 * @param[in]   event   Event generated by button press.
 */
void bsp_event_handler(bsp_event_t event) {
  uint32_t err_code;
  switch (event) {
  case BSP_EVENT_SLEEP:
    sleep_mode_enter();
    break;

  case BSP_EVENT_DISCONNECT:
    err_code = sd_ble_gap_disconnect(m_conn_handle, BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
    if (err_code != NRF_ERROR_INVALID_STATE) {
      APP_ERROR_CHECK(err_code);
    }
    break;

  case BSP_EVENT_WHITELIST_OFF:
    if (m_conn_handle == BLE_CONN_HANDLE_INVALID) {
      err_code = ble_advertising_restart_without_whitelist(&m_advertising);
      if (err_code != NRF_ERROR_INVALID_STATE) {
        APP_ERROR_CHECK(err_code);
      }
    }
    break;

  default:
    break;
  }
}

/**@brief   Function for handling app_uart events.
 *
 * @details This function will receive a single character from the app_uart module and append it to
 *          a string. The string will be be sent over BLE when the last character received was a
 *          'new line' '\n' (hex 0x0A) or if the string has reached the maximum data length.
 */
/**@snippet [Handling the data received over UART] */
void uart_event_handler(void *context, nrf_libuarte_async_evt_t *p_evt) {
  nrf_libuarte_async_t *p_libuarte = (nrf_libuarte_async_t *)context;
  ret_code_t ret;
  uint16_t index = 0;

  switch (p_evt->type) {
  case NRF_LIBUARTE_ASYNC_EVT_ERROR:
    break;
  case NRF_LIBUARTE_ASYNC_EVT_RX_DATA:
    while (index < p_evt->data.rxtx.length) {
      uint16_t length = (uint16_t)(p_evt->data.rxtx.length - index > BLE_NUS_MAX_DATA_LEN) ? BLE_NUS_MAX_DATA_LEN : p_evt->data.rxtx.length - index;
      do {
        ret = ble_nus_data_send(&m_nus, p_evt->data.rxtx.p_data + index, &length, m_conn_handle);
        if ((ret != NRF_ERROR_INVALID_STATE) &&
            (ret != NRF_ERROR_RESOURCES) &&
            (ret != NRF_ERROR_NOT_FOUND)) {
          APP_ERROR_CHECK(ret);
        }
      } while (ret == NRF_ERROR_RESOURCES);
      index += length;
    }
    nrf_libuarte_async_rx_free(p_libuarte, p_evt->data.rxtx.p_data, p_evt->data.rxtx.length);
    break;

  case NRF_LIBUARTE_ASYNC_EVT_TX_DONE:

    break;
  default:
    break;
  }
}
/**@snippet [Handling the data received over UART] */

/**@brief  Function for initializing the UART module.
 */
/**@snippet [UART Initialization] */
static void uart_init(void) {
  uint32_t err_code;

  /* Enable routing of UART pins between the nrf91 SiP and nrf52840
     * https://infocenter.nordicsemi.com/topic/ug_nrf91_dk/UG/nrf91_DK/board_controller.html 
     */
    nrf_gpio_cfg_output(NRF_IF0_2_CTRL);
    nrf_gpio_cfg_output(NRF_IF3_5_CTRL);
    nrf_gpio_pin_set(NRF_IF0_2_CTRL);
    nrf_gpio_pin_set(NRF_IF3_5_CTRL);

    ret_code_t ret = nrf_drv_clock_init();
    APP_ERROR_CHECK(ret);

    nrf_drv_clock_lfclk_request(NULL);
    nrf_drv_clock_hfclk_request(NULL);

    err_code = NRF_LOG_INIT(app_timer_cnt_get);
    //err_code = NRF_LOG_INIT(NULL);
    printf("uart_init: NRF_LOG_INIT\n");
    APP_ERROR_CHECK(err_code);

    NRF_LOG_DEFAULT_BACKENDS_INIT();

/*
    nrf_libuarte_async_config_t nrf_libuarte_async_config = {
      //.tx_pin     = TX_PIN_NUMBER,
      //.rx_pin     = RX_PIN_NUMBER,
      //.tx_pin = 17,
      //.rx_pin = 20,
      .tx_pin     = NRF_GPIO_PIN_MAP(0,17),//TX_PIN_NUMBER,
      .rx_pin     = NRF_GPIO_PIN_MAP(0,20), //RX_PIN_NUMBER,
      .cts_pin    = NRF_GPIO_PIN_MAP(0,15),
      .rts_pin    = NRF_GPIO_PIN_MAP(0,22),
      .baudrate = NRF_UARTE_BAUDRATE_115200,
      .parity = NRF_UARTE_PARITY_EXCLUDED,
      //.hwfc = NRF_UARTE_HWFC_DISABLED,
      .hwfc       = NRF_UARTE_HWFC_ENABLED,
      .timeout_us = 100,
      .int_prio = APP_IRQ_PRIORITY_LOW};
      */

      nrf_libuarte_async_config_t nrf_libuarte_async_config = {
            .tx_pin     = NRF_GPIO_PIN_MAP(0,17),//TX_PIN_NUMBER,
            .rx_pin     = NRF_GPIO_PIN_MAP(0,20), //RX_PIN_NUMBER,
            .cts_pin    = NRF_GPIO_PIN_MAP(0,15),
            .rts_pin    = NRF_GPIO_PIN_MAP(0,22),
            .baudrate   = NRF_UARTE_BAUDRATE_115200,
            .parity     = NRF_UARTE_PARITY_EXCLUDED,
            .hwfc       = NRF_UARTE_HWFC_ENABLED,
            .timeout_us = 100,
            .int_prio   = APP_IRQ_PRIORITY_LOW
    };
      

  err_code = nrf_libuarte_async_init(&libuarte, &nrf_libuarte_async_config, uart_event_handler, (void *)&libuarte);

  APP_ERROR_CHECK(err_code);

  nrf_libuarte_async_enable(&libuarte);

  //static uint8_t text[] = "app_libUARTE example started.\r\n";
  //static uint8_t text_size = sizeof(text);

  //err_code = nrf_libuarte_async_tx(&libuarte, text, text_size);
  //APP_ERROR_CHECK(err_code);
}
/**@snippet [UART Initialization] */


/**@brief Function for initializing buttons and leds.
 *
 * @param[out] p_erase_bonds  Will be true if the clear bonding button was pressed to wake the application up.
 */
static void buttons_leds_init(bool *p_erase_bonds) {
  bsp_event_t startup_event;

  uint32_t err_code = bsp_init(BSP_INIT_LEDS | BSP_INIT_BUTTONS, bsp_event_handler);
  APP_ERROR_CHECK(err_code);

  err_code = bsp_btn_ble_init(NULL, &startup_event);
  APP_ERROR_CHECK(err_code);

  *p_erase_bonds = (startup_event == BSP_EVENT_CLEAR_BONDING_DATA);
}

/**@brief Function for initializing the nrf log module.
 */
static void log_init(void) {
  ret_code_t err_code = NRF_LOG_INIT(NULL);
  APP_ERROR_CHECK(err_code);

  NRF_LOG_DEFAULT_BACKENDS_INIT();
}

/**@brief Function for initializing power management.
 */
static void power_management_init(void) {
  ret_code_t err_code;
  err_code = nrf_pwr_mgmt_init();
  APP_ERROR_CHECK(err_code);
}

/**@brief Function for handling the idle state (main loop).
 *
 * @details If there is no pending log operation, then sleep until next the next event occurs.
 */
static void idle_state_handle(void) {
  if (NRF_LOG_PROCESS() == false) {
    nrf_pwr_mgmt_run();
  }
}


/**@brief Application main function.
 */
int main(void) {
  bool erase_bonds;

  nrf_delay_ms(1000);

  // Initialize.
  uart_init();
  timers_init();
  buttons_leds_init(&erase_bonds);
  power_management_init();
  ant_stack_init();
  ant_channel_rx_broadcast_setup();

  sd_clock_hfclk_request();

  // Start execution.
  NRF_LOG_INFO("Start...");
  // Give time for ant to startup...
  nrf_delay_ms(5000);

  for (;;) {
    static uint8_t text[] = "1";
    static uint8_t text_size = sizeof(text);
    printf("send 1\n");
    nrf_libuarte_async_tx(&libuarte, text, text_size);
    nrf_delay_ms(5000);
    NRF_LOG_FLUSH();
  }


  for (;;) {
    NRF_LOG_FLUSH();
    nrf_pwr_mgmt_run();
    //uart_tx();
    //printf("Waiting\n");
    //nrf_delay_ms(5000);
  }

  // Enter main loop.
  for (;;) {
    idle_state_handle();
  }
}

/**
 * @}
 */